from collections.abc import Mapping, Sequence

from cons.core import ConsError, cons
from etuples import apply as term
from etuples import rands as arguments
from etuples import rator as operator
from unification.core import _reify, _unify, construction_sentinel, reify
from unification.variable import isvar

from .core import eq, lall
from .goals import conso


def applyo(o_rator, o_rands, obj):
    """Construct a goal that relates an object to the application of its (ope)rator to its (ope)rands.

    In other words, this is the relation `op(*args) == obj`.  It uses the
    `rator`, `rands`, and `apply` dispatch functions from `etuples`, so
    implement/override those to get the desired behavior.

    """  # noqa: E501

    def applyo_goal(S):
        nonlocal o_rator, o_rands, obj

        o_rator_rf, o_rands_rf, obj_rf = reify((o_rator, o_rands, obj), S)

        if not isvar(obj_rf):

            # We should be able to use this goal with *any* arguments, so
            # fail when the ground operations fail/err.
            try:
                obj_rator, obj_rands = operator(obj_rf), arguments(obj_rf)
            except (ConsError, NotImplementedError):
                return

            # The object's rator + rands should be the same as the goal's
            yield from lall(eq(o_rator_rf, obj_rator), eq(o_rands_rf, obj_rands))(S)

        elif isvar(o_rands_rf) or isvar(o_rator_rf):
            # The object and at least one of the rand, rators is a logic
            # variable, so let's just assert a `cons` relationship between
            # them
            yield from conso(o_rator_rf, o_rands_rf, obj_rf)(S)
        else:
            # The object is a logic variable, but the rator and rands aren't.
            # We assert that the object is the application of the rand and
            # rators.
            try:
                obj_applied = term(o_rator_rf, o_rands_rf)
            except (ConsError, NotImplementedError):
                return
            yield from eq(obj_rf, obj_applied)(S)

    return applyo_goal


@term.register(object, Sequence)
def term_Sequence(rator, rands):
    # Overwrite the default `apply` dispatch function and make it preserve
    # types
    res = cons(rator, rands)
    return res


def unifiable_with_term(cls):
    _reify.add((cls, Mapping), reify_term)
    _unify.add((cls, cls, Mapping), unify_term)
    return cls


def reify_term(obj, s):
    op, args = operator(obj), arguments(obj)
    op = yield _reify(op, s)
    args = yield _reify(args, s)
    yield construction_sentinel
    yield term(op, args)


def unify_term(u, v, s):
    u_op, u_args = operator(u), arguments(u)
    v_op, v_args = operator(v), arguments(v)
    s = yield _unify(u_op, v_op, s)
    if s is not False:
        s = yield _unify(u_args, v_args, s)
    yield s



"""Functions for associative and commutative unification.

This module provides goals for associative and commutative unification.  It
accomplishes this through naively trying all possibilities.  This was built to
be used in the computer algebra systems SymPy and Theano.

>>> from kanren import run, var, fact
>>> from kanren.assoccomm import eq_assoccomm as eq
>>> from kanren.assoccomm import commutative, associative

>>> # Define some dummy Ops
>>> add = 'add'
>>> mul = 'mul'

>>> # Declare that these ops are commutative using the facts system
>>> fact(commutative, mul)
>>> fact(commutative, add)
>>> fact(associative, mul)
>>> fact(associative, add)

>>> # Define some wild variables
>>> x, y = var('x'), var('y')

>>> # Two expressions to match
>>> pattern = (mul, (add, 1, x), y)                # (1 + x) * y
>>> expr    = (mul, 2, (add, 3, 1))                # 2 * (3 + 1)

>>> print(run(0, (x,y), eq(pattern, expr)))
((3, 2),)
"""
from collections.abc import Sequence
from functools import partial
from operator import eq as equal
from operator import length_hint

from cons.core import ConsPair, car, cdr
from etuples import etuple
from toolz import sliding_window
from unification import reify, unify, var

from .core import conde, eq, ground_order, lall, succeed
from .facts import Relation
from .goals import itero, permuteo
from .graph import term_walko
from .term import term

associative = Relation("associative")
commutative = Relation("commutative")


def flatten_assoc_args(op_predicate, items):
    for i in items:
        if isinstance(i, ConsPair) and op_predicate(car(i)):
            i_cdr = cdr(i)
            if length_hint(i_cdr) > 0:
                yield from flatten_assoc_args(op_predicate, i_cdr)
            else:
                yield i
        else:
            yield i


def assoc_args(rator, rands, n, ctor=None):
    """Produce all associative argument combinations of rator + rands in n-sized rand groupings.

    >>> from kanren.assoccomm import assoc_args
    >>> list(assoc_args('op', [1, 2, 3], 2))
    [[['op', 1, 2], 3], [1, ['op', 2, 3]]]
    """  # noqa: E501
    assert n > 0

    rands_l = list(rands)

    if ctor is None:
        ctor = type(rands)

    if n == len(rands_l):
        yield ctor(rands)
        return

    for i, new_rands in enumerate(sliding_window(n, rands_l)):
        prefix = rands_l[:i]
        new_term = term(rator, ctor(new_rands))
        suffix = rands_l[n + i :]
        res = ctor(prefix + [new_term] + suffix)
        yield res


def eq_assoc_args(
    op, a_args, b_args, n=None, inner_eq=eq, no_ident=False, null_type=etuple
):
    """Create a goal that applies associative unification to an operator and two sets of arguments.

    This is a non-relational utility goal.  It does assumes that the op and at
    least one set of arguments are ground under the state in which it is
    evaluated.
    """  # noqa: E501
    u_args, v_args = var(), var()

    def eq_assoc_args_goal(S):
        nonlocal op, u_args, v_args, n

        (op_rf, u_args_rf, v_args_rf, n_rf) = reify((op, u_args, v_args, n), S)

        if isinstance(v_args_rf, Sequence):
            u_args_rf, v_args_rf = v_args_rf, u_args_rf

        if isinstance(u_args_rf, Sequence) and isinstance(v_args_rf, Sequence):
            # TODO: We just ignore `n` when both are sequences?

            if type(u_args_rf) != type(v_args_rf):
                return

            if no_ident and unify(u_args_rf, v_args_rf, S) is not False:
                return

            op_pred = partial(equal, op_rf)
            u_args_flat = type(u_args_rf)(flatten_assoc_args(op_pred, u_args_rf))
            v_args_flat = type(v_args_rf)(flatten_assoc_args(op_pred, v_args_rf))

            if len(u_args_flat) == len(v_args_flat):
                g = inner_eq(u_args_flat, v_args_flat)
            else:
                if len(u_args_flat) < len(v_args_flat):
                    sm_args, lg_args = u_args_flat, v_args_flat
                else:
                    sm_args, lg_args = v_args_flat, u_args_flat

                grp_sizes = len(lg_args) - len(sm_args) + 1
                assoc_terms = assoc_args(
                    op_rf, lg_args, grp_sizes, ctor=type(u_args_rf)
                )

                g = conde([inner_eq(sm_args, a_args)] for a_args in assoc_terms)

            yield from g(S)

        elif isinstance(u_args_rf, Sequence):
            # TODO: We really need to know the arity (ranges) for the operator
            # in order to make good choices here.
            # For instance, does `(op, 1, 2) == (op, (op, 1, 2))` make sense?
            # If so, the lower-bound on this range should actually be `1`.
            if len(u_args_rf) == 1:
                if not no_ident and (n_rf == 1 or n_rf is None):
                    g = inner_eq(u_args_rf, v_args_rf)
                else:
                    return
            else:

                u_args_flat = list(flatten_assoc_args(partial(equal, op_rf), u_args_rf))

                if n_rf is not None:
                    arg_sizes = [n_rf]
                else:
                    arg_sizes = range(2, len(u_args_flat) + (not no_ident))

                v_ac_args = (
                    v_ac_arg
                    for n_i in arg_sizes
                    for v_ac_arg in assoc_args(
                        op_rf, u_args_flat, n_i, ctor=type(u_args_rf)
                    )
                    if not no_ident or v_ac_arg != u_args_rf
                )
                g = conde([inner_eq(v_args_rf, v_ac_arg)] for v_ac_arg in v_ac_args)

            yield from g(S)

    return lall(
        ground_order((a_args, b_args), (u_args, v_args)),
        itero(u_args, nullo_refs=(v_args,), default_ConsNull=null_type),
        eq_assoc_args_goal,
    )


def eq_assoc(u, v, n=None, op_predicate=associative, null_type=etuple):
    """Create a goal for associative unification of two terms.

    >>> from kanren import run, var, fact
    >>> from kanren.assoccomm import eq_assoc as eq

    >>> fact(commutative, 'add')    # declare that 'add' is commutative
    >>> fact(associative, 'add')    # declare that 'add' is associative

    >>> x = var()
    >>> run(0, x, eq(('add', 1, 2, 3), ('add', 1, x)))
    (('add', 2, 3),)
    """

    def assoc_args_unique(a, b, op, **kwargs):
        return eq_assoc_args(op, a, b, no_ident=True, null_type=null_type)

    return term_walko(op_predicate, assoc_args_unique, u, v, n=n)


def eq_comm(u, v, op_predicate=commutative, null_type=etuple):
    """Create a goal for commutative equality.

    >>> from kanren import run, var, fact
    >>> from kanren.assoccomm import eq_comm as eq
    >>> from kanren.assoccomm import commutative, associative

    >>> fact(commutative, 'add')    # declare that 'add' is commutative
    >>> fact(associative, 'add')    # declare that 'add' is associative

    >>> x = var()
    >>> run(0, x, eq(('add', 1, 2, 3), ('add', 2, x, 1)))
    (3,)
    """

    def permuteo_unique(x, y, op, **kwargs):
        return permuteo(x, y, no_ident=True, default_ConsNull=null_type)

    return term_walko(op_predicate, permuteo_unique, u, v)


def assoc_flatten(a, a_flat):
    def assoc_flatten_goal(S):
        nonlocal a, a_flat

        a_rf = reify(a, S)

        if isinstance(a_rf, Sequence) and (a_rf[0],) in associative.facts:

            def op_pred(sub_op):
                nonlocal S
                sub_op_rf = reify(sub_op, S)
                return sub_op_rf == a_rf[0]

            a_flat_rf = type(a_rf)(flatten_assoc_args(op_pred, a_rf))
        else:
            a_flat_rf = a_rf

        yield from eq(a_flat, a_flat_rf)(S)

    return assoc_flatten_goal


def eq_assoccomm(u, v, null_type=etuple):
    """Construct a goal for associative and commutative unification.

    >>> from kanren.assoccomm import eq_assoccomm as eq
    >>> from kanren.assoccomm import commutative, associative
    >>> from kanren import fact, run, var

    >>> fact(commutative, 'add')    # declare that 'add' is commutative
    >>> fact(associative, 'add')    # declare that 'add' is associative

    >>> x = var()
    >>> e1 = ('add', 1, 2, 3)
    >>> e2 = ('add', 1, x)
    >>> run(0, x, eq(e1, e2))
    (('add', 3, 2), ('add', 2, 3))
    """

    def eq_assoccomm_step(a, b, op):
        z = var()
        return lall(
            # Permute
            conde(
                [
                    commutative(op),
                    permuteo(a, z, no_ident=True, default_ConsNull=etuple),
                ],
                [eq(a, z)],
            ),
            # Generate associative combinations
            conde(
                [associative(op), eq_assoc_args(op, z, b, no_ident=True)], [eq(z, b)]
            ),
        )

    return term_walko(
        lambda x: succeed,
        eq_assoccomm_step,
        u,
        v,
        format_step=assoc_flatten,
        no_ident=False,
    )



# flake8: noqa
"""kanren is a Python library for logic and relational programming."""
from unification import Var, isvar, reify, unifiable, unify, var, variables, vars

from ._version import get_versions
from .core import conde, eq, lall, lany, run
from .facts import Relation, fact, facts
from .goals import (
    appendo,
    conso,
    heado,
    itero,
    membero,
    nullo,
    permuteo,
    permuteq,
    rembero,
    tailo,
)
from .term import arguments, operator, term, unifiable_with_term

__version__ = get_versions()["version"]
del get_versions



from collections.abc import Generator, Sequence
from functools import partial, reduce
from itertools import tee
from operator import length_hint

from cons.core import ConsPair
from toolz import interleave, take
from unification import isvar, reify, unify
from unification.core import isground


def fail(s):
    return iter(())


def succeed(s):
    return iter((s,))


def eq(u, v):
    """Construct a goal stating that its arguments must unify.

    See Also
    --------
        unify
    """

    def eq_goal(s):
        s = unify(u, v, s)
        if s is not False:
            return iter((s,))
        else:
            return iter(())

    return eq_goal


def ldisj_seq(goals):
    """Produce a goal that returns the appended state stream from all successful goal arguments.

    In other words, it behaves like logical disjunction/OR for goals.
    """  # noqa: E501

    if length_hint(goals, -1) == 0:
        return succeed

    def ldisj_seq_goal(S):
        nonlocal goals

        goals, _goals = tee(goals)

        yield from interleave(g(S) for g in _goals)

    return ldisj_seq_goal


def bind(z, g):
    """Apply a goal to a state stream and then combine the resulting state streams."""
    # We could also use `chain`, but `interleave` preserves the old behavior.
    # return chain.from_iterable(map(g, z))
    return interleave(map(g, z))


def lconj_seq(goals):
    """Produce a goal that returns the appended state stream in which all goals are necessarily successful.

    In other words, it behaves like logical conjunction/AND for goals.
    """  # noqa: E501

    if length_hint(goals, -1) == 0:
        return succeed

    def lconj_seq_goal(S):
        nonlocal goals

        goals, _goals = tee(goals)

        g0 = next(_goals, None)

        if g0 is None:
            return

        yield from reduce(bind, _goals, g0(S))

    return lconj_seq_goal


def ldisj(*goals):
    """Form a disjunction of goals."""
    if len(goals) == 1 and isinstance(goals[0], Generator):
        goals = goals[0]

    return ldisj_seq(goals)


def lconj(*goals):
    """Form a conjunction of goals."""
    if len(goals) == 1 and isinstance(goals[0], Generator):
        goals = goals[0]

    return lconj_seq(goals)


def conde(*goals):
    """Form a disjunction of goal conjunctions."""
    if len(goals) == 1 and isinstance(goals[0], Generator):
        goals = goals[0]

    return ldisj_seq(lconj_seq(g) for g in goals)


lall = lconj
lany = ldisj


def ground_order_key(S, x):
    if isvar(x):
        return 2
    elif isground(x, S):
        return -1
    elif issubclass(type(x), ConsPair):
        return 1
    else:
        return 0


def ground_order(in_args, out_args):
    """Construct a non-relational goal that orders a list of terms based on groundedness (grounded precede ungrounded)."""  # noqa: E501

    def ground_order_goal(S):
        nonlocal in_args, out_args

        in_args_rf, out_args_rf = reify((in_args, out_args), S)

        S_new = unify(
            list(out_args_rf) if isinstance(out_args_rf, Sequence) else out_args_rf,
            sorted(in_args_rf, key=partial(ground_order_key, S)),
            S,
        )

        if S_new is not False:
            yield S_new

    return ground_order_goal


def ifa(g1, g2):
    """Create a goal operator that returns the first stream unless it fails."""

    def ifa_goal(S):
        g1_stream = g1(S)
        S_new = next(g1_stream, None)

        if S_new is None:
            yield from g2(S)
        else:
            yield S_new
            yield from g1_stream

    return ifa_goal


def Zzz(gctor, *args, **kwargs):
    """Create an inverse-η-delay for a goal."""

    def Zzz_goal(S):
        yield from gctor(*args, **kwargs)(S)

    return Zzz_goal


def run(n, x, *goals, results_filter=None):
    """Run a logic program and obtain n solutions that satisfy the given goals.

    >>> from kanren import run, var, eq
    >>> x = var()
    >>> run(1, x, eq(x, 1))
    (1,)

    Parameters
    ----------
    n: int
        The number of desired solutions. `n=0` returns a tuple
        with all results and `n=None` returns a lazy sequence of all results.
    x: object
        The form to reify and output.  Usually contains logic variables used in
        the given goals.
    goals: Callables
        A sequence of goals that must be true in logical conjunction
        (i.e. `lall`).
    results_filter: Callable
        A function to apply to the results stream (e.g. a `unique` filter).
    """
    results = map(partial(reify, x), lall(*goals)({}))

    if results_filter is not None:
        results = results_filter(results)

    if n is None:
        return results
    elif n == 0:
        return tuple(results)
    else:
        return tuple(take(n, results))


def dbgo(*args, msg=None):  # pragma: no cover
    """Construct a goal that sets a debug trace and prints reified arguments."""
    from pprint import pprint

    def dbgo_goal(S):
        nonlocal args
        args = reify(args, S)

        if msg is not None:
            print(msg)

        pprint(args)

        import pdb

        pdb.set_trace()
        yield S

    return dbgo_goal



from collections import namedtuple
from collections.abc import Hashable, Iterable, Mapping, MutableSet, Set
from itertools import chain

HashableForm = namedtuple("HashableForm", ["type", "data"])


class FlexibleSet(MutableSet):
    """A set that uses a list (and costly identity check) for unhashable items."""

    __slots__ = ("set", "list")

    def __init__(self, iterable=None):

        self.set = set()
        self.list = []

        if iterable is not None:
            for i in iterable:
                self.add(i)

    def add(self, item):
        try:
            self.set.add(item)
        except TypeError:
            # TODO: Could try `make_hashable`.
            # TODO: Use `bisect` for unhashable but orderable elements
            if item not in self.list:
                self.list.append(item)

    def discard(self, item):
        try:
            self.remove(item)
        except KeyError:
            pass

    def clear(self):
        self.set.clear()
        self.list.clear()

    def pop(self):
        try:
            return self.set.pop()
        except (TypeError, KeyError):
            try:
                return self.list.pop(-1)
            except IndexError:
                raise KeyError()

    def remove(self, item):
        try:
            self.set.remove(item)
        except (TypeError, KeyError):
            try:
                self.list.remove(item)
            except ValueError:
                raise KeyError()

    def copy(self):
        res = type(self)()
        res.set = self.set.copy()
        res.list = self.list.copy()
        return res

    def __le__(self, other):
        raise NotImplementedError()

    def __ge__(self, other):
        raise NotImplementedError()

    def __ior__(self, item):
        raise NotImplementedError()

    def __iand__(self, item):
        raise NotImplementedError()

    def __ixor__(self, item):
        raise NotImplementedError()

    def __isub__(self, item):
        raise NotImplementedError()

    def __iter__(self):
        return chain(self.set, self.list)

    def __contains__(self, value):
        try:
            return value in self.set or value in self.list
        except TypeError:
            return value in self.list

    def __len__(self):
        return len(self.set) + len(self.list)

    def __eq__(self, other):
        if type(self) == type(other):
            return self.set == other.set and self.list == other.list
        elif isinstance(other, Set):
            return len(self.list) == 0 and other.issuperset(self.set)

        return NotImplemented

    def __repr__(self):
        return f"FlexibleSet([{', '.join(str(s) for s in self)}])"


def hashable(x):
    try:
        hash(x)
        return True
    except TypeError:
        return False


def dicthash(d):
    return hash(frozenset(d.items()))


def make_hashable(x):
    # TODO: Better as a dispatch function?
    if hashable(x):
        return x
    if isinstance(x, slice):
        return HashableForm(type(x), (x.start, x.stop, x.step))
    if isinstance(x, Mapping):
        return HashableForm(type(x), frozenset(tuple(multihash(i) for i in x.items())))
    if isinstance(x, Iterable):
        return HashableForm(type(x), tuple(multihash(i) for i in x))
    raise TypeError(f"Hashing not covered for {x}")


def multihash(x):
    return hash(make_hashable(x))


def unique(seq, key=lambda x: x):
    seen = set()
    for item in seq:
        try:
            k = key(item)
        except TypeError:
            # Just yield it and hope for the best, since we can't efficiently
            # check if we've seen it before.
            yield item
            continue
        if not isinstance(k, Hashable):
            # Just yield it and hope for the best, since we can't efficiently
            # check if we've seen it before.
            yield item
        elif k not in seen:
            seen.add(key(item))
            yield item


def intersection(*seqs):
    return (item for item in seqs[0] if all(item in seq for seq in seqs[1:]))


def groupsizes(total, len):
    """Construct groups of length len that add up to total.

    >>> from kanren.util import groupsizes
    >>> tuple(groupsizes(4, 2))
    ((1, 3), (2, 2), (3, 1))
    """
    if len == 1:
        yield (total,)
    else:
        for i in range(1, total - len + 1 + 1):
            for perm in groupsizes(total - i, len - 1):
                yield (i,) + perm


def pprint(g):  # pragma: no cover
    """Pretty print a tree of goals."""
    if callable(g) and hasattr(g, "__name__"):
        return g.__name__
    if isinstance(g, type):
        return g.__name__
    if isinstance(g, tuple):
        return "(" + ", ".join(map(pprint, g)) + ")"
    return str(g)


def index(tup, ind):
    """Fancy indexing with tuples."""
    return tuple(tup[i] for i in ind)



from functools import partial

from etuples import etuple
from unification import isvar, reify, var

from .core import Zzz, conde, eq, fail, ground_order, lall, succeed
from .goals import conso, nullo
from .term import applyo


def mapo(relation, a, b, null_type=list, null_res=True, first=True):
    """Apply a relation to corresponding elements in two sequences and succeed if the relation succeeds for all pairs."""  # noqa: E501

    b_car, b_cdr = var(), var()
    a_car, a_cdr = var(), var()

    return conde(
        [nullo(a, b, default_ConsNull=null_type) if (not first or null_res) else fail],
        [
            conso(a_car, a_cdr, a),
            conso(b_car, b_cdr, b),
            Zzz(relation, a_car, b_car),
            Zzz(mapo, relation, a_cdr, b_cdr, null_type=null_type, first=False),
        ],
    )


def map_anyo(
    relation, a, b, null_type=list, null_res=False, first=True, any_succeed=False
):
    """Apply a relation to corresponding elements in two sequences and succeed if at least one pair succeeds.

    Parameters
    ----------
    null_type: optional
       An object that's a valid cdr for the collection type desired.  If
       `False` (i.e. the default value), the cdr will be inferred from the
       inputs, or defaults to an empty list.
    """  # noqa: E501

    b_car, b_cdr = var(), var()
    a_car, a_cdr = var(), var()

    return conde(
        [
            nullo(a, b, default_ConsNull=null_type)
            if (any_succeed or (first and null_res))
            else fail
        ],
        [
            conso(a_car, a_cdr, a),
            conso(b_car, b_cdr, b),
            conde(
                [
                    Zzz(relation, a_car, b_car),
                    Zzz(
                        map_anyo,
                        relation,
                        a_cdr,
                        b_cdr,
                        null_type=null_type,
                        any_succeed=True,
                        first=False,
                    ),
                ],
                [
                    eq(a_car, b_car),
                    Zzz(
                        map_anyo,
                        relation,
                        a_cdr,
                        b_cdr,
                        null_type=null_type,
                        any_succeed=any_succeed,
                        first=False,
                    ),
                ],
            ),
        ],
    )


def vararg_success(*args):
    return succeed


def eq_length(u, v, default_ConsNull=list):
    """Construct a goal stating that two sequences are the same length and type."""

    return mapo(vararg_success, u, v, null_type=default_ConsNull)


def reduceo(relation, in_term, out_term, *args, **kwargs):
    """Relate a term and the fixed-point of that term under a given relation.

    This includes the "identity" relation.
    """

    def reduceo_goal(s):

        nonlocal in_term, out_term, relation, args, kwargs

        in_term_rf, out_term_rf = reify((in_term, out_term), s)

        # The result of reducing the input graph once
        term_rdcd = var()

        # Are we working "backward" and (potentially) "expanding" a graph
        # (e.g. when the relation is a reduction rule)?
        is_expanding = isvar(in_term_rf)

        # One application of the relation assigned to `term_rdcd`
        single_apply_g = relation(in_term_rf, term_rdcd, *args, **kwargs)

        # Assign/equate (unify, really) the result of a single application to
        # the "output" term.
        single_res_g = eq(term_rdcd, out_term_rf)

        # Recurse into applications of the relation (well, produce a goal that
        # will do that)
        another_apply_g = reduceo(relation, term_rdcd, out_term_rf, *args, **kwargs)

        # We want the fixed-point value to show up in the stream output
        # *first*, but that requires some checks.
        if is_expanding:
            # When an un-reduced term is a logic variable (e.g. we're
            # "expanding"), we can't go depth first.
            # We need to draw the association between (i.e. unify) the reduced
            # and expanded terms ASAP, in order to produce finite
            # expanded graphs first and yield results.
            #
            # In other words, there's no fixed-point to produce in this
            # situation.  Instead, for example, we have to produce an infinite
            # stream of terms that have `out_term_rf` as a fixed point.
            # g = conde([single_res_g, single_apply_g],
            #           [another_apply_g, single_apply_g])
            g = lall(conde([single_res_g], [another_apply_g]), single_apply_g)
        else:
            # Run the recursion step first, so that we get the fixed-point as
            # the first result
            g = lall(single_apply_g, conde([another_apply_g], [single_res_g]))

        yield from g(s)

    return reduceo_goal


def walko(
    goal,
    graph_in,
    graph_out,
    rator_goal=None,
    null_type=etuple,
    map_rel=partial(map_anyo, null_res=True),
):
    """Apply a binary relation between all nodes in two graphs.

    When `rator_goal` is used, the graphs are treated as term graphs, and the
    multi-functions `rator`, `rands`, and `apply` are used to walk the graphs.
    Otherwise, the graphs must be iterable according to `map_anyo`.

    Parameters
    ----------
    goal: callable
        A goal that is applied to all terms in the graph.
    graph_in: object
        The graph for which the left-hand side of a binary relation holds.
    graph_out: object
        The graph for which the right-hand side of a binary relation holds.
    rator_goal: callable (default None)
        A goal that is applied to the rators of a graph.  When specified,
        `goal` is only applied to rands and it must succeed along with the
        rator goal in order to descend into sub-terms.
    null_type: type
        The collection type used when it is not fully determined by the graph
        arguments.
    map_rel: callable
        The map relation used to apply `goal` to a sub-graph.
    """

    def walko_goal(s):

        nonlocal goal, rator_goal, graph_in, graph_out, null_type, map_rel

        graph_in_rf, graph_out_rf = reify((graph_in, graph_out), s)

        rator_in, rands_in, rator_out, rands_out = var(), var(), var(), var()

        _walko = partial(
            walko, goal, rator_goal=rator_goal, null_type=null_type, map_rel=map_rel
        )

        g = conde(
            # TODO: Use `Zzz`, if needed.
            [
                goal(graph_in_rf, graph_out_rf),
            ],
            [
                lall(
                    applyo(rator_in, rands_in, graph_in_rf),
                    applyo(rator_out, rands_out, graph_out_rf),
                    rator_goal(rator_in, rator_out),
                    map_rel(_walko, rands_in, rands_out, null_type=null_type),
                )
                if rator_goal is not None
                else map_rel(_walko, graph_in_rf, graph_out_rf, null_type=null_type),
            ],
        )

        yield from g(s)

    return walko_goal


def term_walko(
    rator_goal,
    rands_goal,
    a,
    b,
    null_type=etuple,
    no_ident=False,
    format_step=None,
    **kwargs
):
    """Construct a goal for walking a term graph.

    This implementation is somewhat specific to the needs of `eq_comm` and
    `eq_assoc`, but it could be transferred to `kanren.graph`.

    XXX: Make sure `rator_goal` will succeed for unground logic variables;
    otherwise, this will diverge.
    XXX: `rands_goal` should not be contain `eq`, i.e. `rands_goal(x, x)`
    should always fail!
    """

    def single_step(s, t):
        u, v = var(), var()
        u_rator, u_rands = var(), var()
        v_rands = var()

        return lall(
            ground_order((s, t), (u, v)),
            applyo(u_rator, u_rands, u),
            applyo(u_rator, v_rands, v),
            rator_goal(u_rator),
            # These make sure that there are at least two rands, which
            # makes sense for commutativity and associativity, at least.
            conso(var(), var(), u_rands),
            conso(var(), var(), v_rands),
            Zzz(rands_goal, u_rands, v_rands, u_rator, **kwargs),
        )

    def term_walko_step(s, t):
        nonlocal rator_goal, rands_goal, null_type
        u, v = var(), var()
        z, w = var(), var()

        return lall(
            ground_order((s, t), (u, v)),
            format_step(u, w) if format_step is not None else eq(u, w),
            conde(
                [
                    # Apply, then walk or return
                    single_step(w, v),
                ],
                [
                    # Walk, then apply or return
                    map_anyo(term_walko_step, w, z, null_type=null_type),
                    conde([eq(z, v)], [single_step(z, v)]),
                ],
            ),
        )

    return lall(
        term_walko_step(a, b)
        if no_ident
        else conde([term_walko_step(a, b)], [eq(a, b)]),
    )



import weakref
from abc import ABC, abstractmethod
from collections import UserDict
from collections.abc import Mapping

from cons.core import ConsPair
from toolz import groupby
from unification import Var, reify, unify, var
from unification.core import _reify, isground
from unification.utils import transitive_get as walk

from .util import FlexibleSet


class ConstraintStore(ABC):
    """A class that enforces constraints between logic variables in a miniKanren state.

    Attributes
    ----------
    lvar_constraints: MutableMapping
        A mapping of logic variables to sets of objects that define their
        constraints (e.g. a set of items with which the logic variable cannot
        be unified).  The mapping's values are entirely determined by the
        ConstraintStore implementation.

    """

    __slots__ = ("lvar_constraints",)
    op_str = None

    def __init__(self, lvar_constraints=None):
        # self.lvar_constraints = weakref.WeakKeyDictionary(lvar_constraints)
        self.lvar_constraints = lvar_constraints or dict()

    @abstractmethod
    def pre_unify_check(self, lvar_map, lvar=None, value=None):
        """Check a key-value pair before they're added to a ConstrainedState."""
        raise NotImplementedError()

    @abstractmethod
    def post_unify_check(self, lvar_map, lvar=None, value=None, old_state=None):
        """Check a key-value pair after they're added to a ConstrainedState.

        XXX: This method may alter the internal constraints, so make a copy!
        """
        raise NotImplementedError()

    def add(self, lvar, lvar_constraint, **kwargs):
        """Add a new constraint."""
        if lvar not in self.lvar_constraints:
            self.lvar_constraints[lvar] = FlexibleSet([lvar_constraint])
        else:
            self.lvar_constraints[lvar].add(lvar_constraint)

    def constraints_str(self, lvar):
        """Print the constraints on a logic variable."""
        if lvar in self.lvar_constraints:
            return f"{self.op_str} {self.lvar_constraints[lvar]}"
        else:
            return ""

    def copy(self):
        return type(self)(
            lvar_constraints={k: v.copy() for k, v in self.lvar_constraints.items()},
        )

    def __contains__(self, lvar):
        return lvar in self.lvar_constraints

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.op_str == other.op_str
            and self.lvar_constraints == other.lvar_constraints
        )

    def __repr__(self):
        return f"ConstraintStore({self.op_str}: {self.lvar_constraints})"


class ConstrainedState(UserDict):
    """A miniKanren state that holds unifications of logic variables and upholds constraints on logic variables."""  # noqa: E501

    __slots__ = ("constraints",)

    def __init__(self, *s, constraints=None):
        super().__init__(*s)
        self.constraints = dict(constraints or [])

    def pre_unify_checks(self, lvar, value):
        """Check the constraints before unification."""
        return all(
            cstore.pre_unify_check(self.data, lvar, value)
            for cstore in self.constraints.values()
        )

    def post_unify_checks(self, lvar_map, lvar, value):
        """Check constraints and return an updated state and constraints.

        Returns
        -------
        A new `ConstrainedState` and `False`.

        """
        S = self.copy(data=lvar_map)
        if any(
            not cstore.post_unify_check(lvar_map, lvar, value, old_state=S)
            for cstore in S.constraints.values()
        ):
            return False

        return S

    def copy(self, data=None):
        if data is None:
            data = self.data.copy()
        return type(self)(
            data, constraints={k: v.copy() for k, v in self.constraints.items()}
        )

    def __eq__(self, other):
        if isinstance(other, ConstrainedState):
            return self.data == other.data and self.constraints == other.constraints

        if isinstance(other, Mapping) and not self.constraints:
            return self.data == other

        return False

    def __repr__(self):
        return f"ConstrainedState({repr(self.data)}, {self.constraints})"


def unify_ConstrainedState(u, v, S):
    if S.pre_unify_checks(u, v):
        s = unify(u, v, S.data)
        if s is not False:
            S = S.post_unify_checks(s, u, v)
            if S is not False:
                return S

    return False


unify.add((object, object, ConstrainedState), unify_ConstrainedState)


class ConstrainedVar(Var):
    """A logic variable that tracks its own constraints.

    Currently, this is only for display/reification purposes.

    """

    __slots__ = ("S", "var")

    def __init__(self, var, S):
        self.S = weakref.ref(S)
        self.token = var.token
        self.var = weakref.ref(var)

    def __repr__(self):
        S = self.S()
        var = self.var()
        res = super().__repr__()
        if S is not None and var is not None:
            u_constraints = ",".join(
                [c.constraints_str(var) for c in S.constraints.values()]
            )
            return f"{res}: {{{u_constraints}}}"
        else:
            return res

    def __eq__(self, other):
        if type(other) == type(self):
            return self.S == other.S and self.token == other.token
        elif type(other) == Var:
            # NOTE: A more valid comparison is same token and no constraints.
            return self.token == other.token
        return NotImplemented

    def __hash__(self):
        return hash((Var, self.token))


def _reify_ConstrainedState(u, S):
    u_res = walk(u, S.data)

    if u_res is u:
        yield ConstrainedVar(u_res, S)
    else:
        yield _reify(u_res, S)


_reify.add((Var, ConstrainedState), _reify_ConstrainedState)


class DisequalityStore(ConstraintStore):
    """A disequality constraint (i.e. two things do not unify)."""

    op_str = "neq"

    def __init__(self, lvar_constraints=None):
        super().__init__(lvar_constraints)

    def post_unify_check(self, lvar_map, lvar=None, value=None, old_state=None):

        for lv_key, constraints in list(self.lvar_constraints.items()):
            lv = reify(lv_key, lvar_map)
            constraints_rf = reify(tuple(constraints), lvar_map)

            for cs in constraints_rf:
                s = unify(lv, cs, {})

                if s is not False and not s:
                    # They already unify, but with no unground logic variables,
                    # so we have an immediate violation of the constraint.
                    return False
                elif s is False:
                    # They don't unify and have no unground logic variables, so
                    # the constraint is immediately satisfied and there's no
                    # reason to continue checking this constraint.
                    constraints.discard(cs)
                else:
                    # They unify when/if the unifications in `s` are made, so
                    # let's add these as new constraints.
                    for k, v in s.items():
                        self.add(k, v)

            if len(constraints) == 0:
                # This logic variable has no more unground constraints, so
                # remove it.
                del self.lvar_constraints[lv_key]

        return True

    def pre_unify_check(self, lvar_map, lvar=None, value=None):
        return True


def neq(u, v):
    """Construct a disequality goal."""

    def neq_goal(S):
        nonlocal u, v

        u_rf, v_rf = reify((u, v), S)

        # Get the unground logic variables that would unify the two objects;
        # these are all the logic variables that we can't let unify.
        s_uv = unify(u_rf, v_rf, {})

        if s_uv is False:
            # They don't unify and have no unground logic variables, so the
            # constraint is immediately satisfied.
            yield S
            return
        elif not s_uv:
            # They already unify, but with no unground logic variables, so we
            # have an immediate violation of the constraint.
            return

        if not isinstance(S, ConstrainedState):
            S = ConstrainedState(S)

        cs = S.constraints.setdefault(DisequalityStore, DisequalityStore())

        for lvar, obj in s_uv.items():
            cs.add(lvar, obj)

        # We need to check the current state for validity.
        if cs.post_unify_check(S.data):
            yield S

    return neq_goal


class PredicateStore(ConstraintStore, ABC):
    """An abstract store for testing simple predicates."""

    # Require that all constraints be satisfied for a term; otherwise, succeed
    # if only one is satisfied.
    require_all_constraints = True

    # @abstractmethod
    # def cterm_type_check(self, lvt):
    #     """Check the type of the constrained term when it's ground."""
    #     raise NotImplementedError()

    @abstractmethod
    def cparam_type_check(self, lvt):
        """Check the type of the constraint parameter when it's ground."""
        raise NotImplementedError()

    @abstractmethod
    def constraint_check(self, lv, lvt):
        """Check the constrained term against the constraint parameters when they're ground.

        I.e. test the constraint.
        """
        raise NotImplementedError()

    @abstractmethod
    def constraint_isground(self, lv, lvar_map):
        """Check whether or not the constrained term is "ground enough" to be checked."""  # noqa: E501
        raise NotImplementedError()

    def post_unify_check(self, lvar_map, lvar=None, value=None, old_state=None):

        for lv_key, constraints in list(self.lvar_constraints.items()):

            lv = reify(lv_key, lvar_map)

            is_lv_ground = self.constraint_isground(lv, lvar_map) or isground(
                lv, lvar_map
            )

            if not is_lv_ground:
                # This constraint isn't ready to be checked
                continue

            # if is_lv_ground and not self.cterm_type_check(lv):
            #     self.lvar_constraints[lv_key]
            #     return False

            constraint_grps = groupby(
                lambda x: isground(x, lvar_map), reify(iter(constraints), lvar_map)
            )

            constraints_unground = constraint_grps.get(False, ())
            constraints_ground = constraint_grps.get(True, ())

            if len(constraints_ground) > 0 and not all(
                self.cparam_type_check(c) for c in constraints_ground
            ):
                # Some constraint parameters aren't the correct type, so fail.
                # del self.lvar_constraints[lv_key]
                return False

            assert constraints_unground or constraints_ground

            if is_lv_ground and len(constraints_unground) == 0:

                if self.require_all_constraints and any(
                    not self.constraint_check(lv, t) for t in constraints_ground
                ):
                    return False
                elif not self.require_all_constraints and not any(
                    self.constraint_check(lv, t) for t in constraints_ground
                ):
                    return False

                # The instance and constraint parameters are all ground and the
                # constraint is satisfied, so, since nothing should change from
                # here on, we can remove the constraint.

                del self.lvar_constraints[lv_key]

        # Some types are unground, so we continue checking until they are
        return True

    def pre_unify_check(self, lvar_map, lvar=None, value=None):
        return True


class TypeStore(PredicateStore):
    """A constraint store for asserting object types."""

    require_all_constraints = True

    op_str = "typeo"

    def __init__(self, lvar_constraints=None):
        super().__init__(lvar_constraints)

    def add(self, lvt, cparams):
        if lvt in self.lvar_constraints:
            raise ValueError("Only one type constraint can be applied to a term")

        return super().add(lvt, cparams)

    # def cterm_type_check(self, lvt):
    #     return True

    def cparam_type_check(self, x):
        return isinstance(x, type)

    def constraint_check(self, x, cx):
        return type(x) == cx

    def constraint_isground(self, lv, lvar_map):
        return not (isinstance(lv, Var) or issubclass(type(lv), ConsPair))


def typeo(u, u_type):
    """Construct a goal specifying the type of a term."""

    def typeo_goal(S):
        nonlocal u, u_type

        u_rf, u_type_rf = reify((u, u_type), S)

        if not isground(u_rf, S) or not isground(u_type_rf, S):

            if not isinstance(S, ConstrainedState):
                S = ConstrainedState(S)

            cs = S.constraints.setdefault(TypeStore, TypeStore())

            try:
                cs.add(u_rf, u_type_rf)
            except TypeError:
                # If the instance object can't be hashed, we can simply use a
                # logic variable to uniquely identify it.
                u_lv = var()
                S[u_lv] = u_rf
                cs.add(u_lv, u_type_rf)

            if cs.post_unify_check(S.data, u_rf, u_type_rf):
                yield S

        elif isinstance(u_type_rf, type) and type(u_rf) == u_type_rf:
            yield S

    return typeo_goal


class IsinstanceStore(PredicateStore):
    """A constraint store for asserting object instance types."""

    op_str = "isinstanceo"

    # Satisfying any one constraint is good enough
    require_all_constraints = False

    def __init__(self, lvar_constraints=None):
        super().__init__(lvar_constraints)

    # def cterm_type_check(self, lvt):
    #     return True

    def cparam_type_check(self, lvt):
        return isinstance(lvt, type)

    def constraint_check(self, lv, lvt):
        return isinstance(lv, lvt)

    def constraint_isground(self, lv, lvar_map):
        return not (isinstance(lv, Var) or issubclass(type(lv), ConsPair))


def isinstanceo(u, u_type):
    """Construct a goal specifying that a term is an instance of a type.

    Only a single instance type can be assigned per goal, i.e.

        lany(isinstanceo(var(), list),
             isinstanceo(var(), tuple))

    and not

        isinstanceo(var(), (list, tuple))

    """

    def isinstanceo_goal(S):
        nonlocal u, u_type

        u_rf, u_type_rf = reify((u, u_type), S)

        if not isground(u_rf, S) or not isground(u_type_rf, S):

            if not isinstance(S, ConstrainedState):
                S = ConstrainedState(S)

            cs = S.constraints.setdefault(IsinstanceStore, IsinstanceStore())

            try:
                cs.add(u_rf, u_type_rf)
            except TypeError:
                # If the instance object can't be hashed, we can simply use a
                # logic variable to uniquely identify it.
                u_lv = var()
                S[u_lv] = u_rf
                cs.add(u_lv, u_type_rf)

            if cs.post_unify_check(S.data, u_rf, u_type_rf):
                yield S

        # elif isground(u_type, S):
        #     yield from lany(eq(u_type, u_t) for u_t in type(u).mro())(S)
        elif (
            isinstance(u_type_rf, type)
            # or (
            #     isinstance(u_type, Iterable)
            #     and all(isinstance(t, type) for t in u_type)
            # )
        ) and isinstance(u_rf, u_type_rf):
            yield S

    return isinstanceo_goal



from toolz import merge
from unification import reify, unify

from .util import intersection


class Relation(object):
    _id = 0

    def __init__(self, name=None):
        self.facts = set()
        self.index = dict()
        if not name:
            name = "_%d" % Relation._id
            Relation._id += 1
        self.name = name

    def add_fact(self, *inputs):
        """Add a fact to the knowledge-base.

        See Also
        --------
            fact
            facts
        """
        fact = tuple(inputs)

        self.facts.add(fact)

        for key in enumerate(inputs):
            if key not in self.index:
                self.index[key] = set()
            self.index[key].add(fact)

    def __call__(self, *args):
        """Return a goal that produces a list of substitutions matching a fact in the knowledge-base.

        >>> from kanren.facts import Relation
        >>> from unification import var
        >>>
        >>> x, y = var('x'), var('y')
        >>> r = Relation()
        >>> r.add_fact(1, 2, 3)
        >>> r.add_fact(4, 5, 6)
        >>> list(r(x, y, 3)({})) == [{y: 2, x: 1}]
        True
        >>> list(r(x, 5, y)({})) == [{y: 6, x: 4}]
        True
        >>> list(r(x, 42, y)({}))
        []

        Parameters
        ----------
        *args:
            The goal to evaluate. This consists of vars and values to match
            facts against.

        """  # noqa: E501

        def goal(substitution):
            args2 = reify(args, substitution)
            subsets = [self.index[key] for key in enumerate(args) if key in self.index]
            if subsets:  # we are able to reduce the pool early
                facts = intersection(*sorted(subsets, key=len))
            else:
                facts = self.facts

            for fact in facts:
                unified = unify(fact, args2, substitution)
                if unified != False:
                    yield merge(unified, substitution)

        return goal

    def __str__(self):
        return "Rel: " + self.name

    __repr__ = __str__


def fact(rel, *args):
    """Declare a fact.

    >>> from kanren import fact, Relation, var, run
    >>> parent = Relation()
    >>> fact(parent, "Homer", "Bart")
    >>> fact(parent, "Homer", "Lisa")

    >>> x = var()
    >>> run(1, x, parent(x, "Bart"))
    ('Homer',)
    """
    rel.add_fact(*args)


def facts(rel, *lists):
    """Declare several facts.

    >>> from kanren import fact, Relation, var, run
    >>> parent = Relation()
    >>> facts(parent,  ("Homer", "Bart"),
    ...                ("Homer", "Lisa"))

    >>> x = var()
    >>> run(1, x, parent(x, "Bart"))
    ('Homer',)
    """
    for lst in lists:
        fact(rel, *lst)



from collections import Counter
from collections.abc import Sequence
from functools import partial
from itertools import permutations
from operator import length_hint

from cons import cons
from cons.core import ConsNull, ConsPair
from unification import reify, var
from unification.core import isground

from .core import conde, eq, lall, lany


def heado(head, coll):
    """Construct a goal stating that head is the head of coll.

    See Also
    --------
        tailo
        conso
    """
    return eq(cons(head, var()), coll)


def tailo(tail, coll):
    """Construct a goal stating that tail is the tail of coll.

    See Also
    --------
        heado
        conso
    """
    return eq(cons(var(), tail), coll)


def conso(h, t, r):
    """Construct a goal stating that cons h + t == l."""
    return eq(cons(h, t), r)


def nullo(*args, refs=None, default_ConsNull=list):
    """Create a goal asserting that one or more terms are a/the same `ConsNull` type.

    `ConsNull` types return proper Python collections when used as a CDR value
    in a CONS (e.g. `cons(1, []) == [1]`).

    This goal doesn't require that all args be unifiable; only that they have
    the same `ConsNull` type.  Unlike the classic `lall(eq(x, []), eq(y, x))`
    `conde`-branch idiom used when recursively walking a single sequence via
    `conso`, this allows us to perform the same essential function while
    walking distinct lists that do not necessarily terminate on the same
    iteration.

    Parameters
    ----------
    args: tuple of objects
        The terms to consider as an instance of the `ConsNull` type
    refs: tuple of objects
        The terms to use as reference types.  These are not unified with the
        `ConsNull` type, instead they are used to constrain the `ConsNull`
        types considered valid.
    default_ConsNull: type
        The sequence type to use when all logic variables are unground.

    """

    def nullo_goal(s):

        nonlocal args, default_ConsNull

        if refs is not None:
            refs_rf = reify(refs, s)
        else:
            refs_rf = ()

        args_rf = reify(args, s)

        arg_null_types = set(
            # Get an empty instance of the type
            type(a)
            for a in args_rf + refs_rf
            # `ConsPair` and `ConsNull` types that are not literally `ConsPair`s
            if isinstance(a, (ConsPair, ConsNull)) and not issubclass(type(a), ConsPair)
        )

        try:
            null_type = arg_null_types.pop()
        except KeyError:
            null_type = default_ConsNull

        if len(arg_null_types) > 0 and any(a != null_type for a in arg_null_types):
            # Mismatching null types: fail.
            return

        g = lall(*[eq(a, null_type()) for a in args_rf])

        yield from g(s)

    return nullo_goal


def itero(lst, nullo_refs=None, default_ConsNull=list):
    """Construct a goal asserting that a term is an iterable type.

    This is a generic version of the standard `listo` that accounts for
    different iterable types supported by `cons` in Python.

    See `nullo`
    """

    def itero_goal(S):
        nonlocal lst, nullo_refs, default_ConsNull
        l_rf = reify(lst, S)
        c, d = var(), var()
        g = conde(
            [nullo(l_rf, refs=nullo_refs, default_ConsNull=default_ConsNull)],
            [conso(c, d, l_rf), itero(d, default_ConsNull=default_ConsNull)],
        )
        yield from g(S)

    return itero_goal


def membero(x, ls):
    """Construct a goal stating that x is an item of coll."""

    def membero_goal(S):
        nonlocal x, ls

        x_rf, ls_rf = reify((x, ls), S)
        a, d = var(), var()

        g = lall(conso(a, d, ls), conde([eq(a, x)], [membero(x, d)]))

        yield from g(S)

    return membero_goal


def appendo(lst, s, out, default_ConsNull=list):
    """Construct a goal for the relation lst + s = ls.

    See Byrd thesis pg. 247
    https://scholarworks.iu.edu/dspace/bitstream/handle/2022/8777/Byrd_indiana_0093A_10344.pdf
    """

    def appendo_goal(S):
        nonlocal lst, s, out

        l_rf, s_rf, out_rf = reify((lst, s, out), S)

        a, d, res = var(prefix="a"), var(prefix="d"), var(prefix="res")

        _nullo = partial(nullo, default_ConsNull=default_ConsNull)

        g = conde(
            [
                # All empty
                _nullo(s_rf, l_rf, out_rf),
            ],
            [
                # `lst` is empty
                conso(a, d, out_rf),
                eq(s_rf, out_rf),
                _nullo(l_rf, refs=(s_rf, out_rf)),
            ],
            [
                conso(a, d, l_rf),
                conso(a, res, out_rf),
                appendo(d, s_rf, res, default_ConsNull=default_ConsNull),
            ],
        )

        yield from g(S)

    return appendo_goal


def rembero(x, lst, o, default_ConsNull=list):
    """Remove the first occurrence of `x` in `lst` resulting in `o`."""

    from .constraints import neq

    def rembero_goal(s):
        nonlocal x, lst, o

        x_rf, l_rf, o_rf = reify((x, lst, o), s)

        l_car, l_cdr, r = var(), var(), var()

        g = conde(
            [
                nullo(l_rf, o_rf, default_ConsNull=default_ConsNull),
            ],
            [
                conso(l_car, l_cdr, l_rf),
                eq(x_rf, l_car),
                eq(l_cdr, o_rf),
            ],
            [
                conso(l_car, l_cdr, l_rf),
                neq(l_car, x),
                conso(l_car, r, o_rf),
                rembero(x_rf, l_cdr, r, default_ConsNull=default_ConsNull),
            ],
        )

        yield from g(s)

    return rembero_goal


def permuteo(a, b, inner_eq=eq, default_ConsNull=list, no_ident=False):
    """Construct a goal asserting equality or sequences under permutation.

    For example, (1, 2, 2) equates to (2, 1, 2) under permutation
    >>> from kanren import var, run, permuteo
    >>> x = var()
    >>> run(0, x, permuteo(x, (1, 2)))
    ((1, 2), (2, 1))

    >>> run(0, x, permuteo((2, 1, x), (2, 1, 2)))
    (2,)
    """

    def permuteo_goal(S):
        nonlocal a, b, default_ConsNull, inner_eq

        a_rf, b_rf = reify((a, b), S)

        # If the lengths differ, then fail
        a_len, b_len = length_hint(a_rf, -1), length_hint(b_rf, -1)
        if a_len > 0 and b_len > 0 and a_len != b_len:
            return

        if isinstance(a_rf, Sequence):

            a_type = type(a_rf)

            a_perms = permutations(a_rf)

            if no_ident:
                next(a_perms)

            if isinstance(b_rf, Sequence):

                b_type = type(b_rf)

                # Fail on mismatched types or straight equality (when
                # `no_ident` is enabled)
                if a_type != b_type or (no_ident and a_rf == b_rf):
                    return

                try:
                    # `a` and `b` are sequences, so let's see if we can pull out
                    # all the (hash-)equivalent elements.
                    # XXX: Use of this requires that the equivalence relation
                    # implied by `inner_eq` be a *superset* of `eq`.

                    cntr_a, cntr_b = Counter(a_rf), Counter(b_rf)
                    rdcd_a, rdcd_b = cntr_a - cntr_b, cntr_b - cntr_a

                    if len(rdcd_a) == len(rdcd_b) == 0:
                        yield S
                        return
                    elif len(rdcd_a) < len(cntr_a):
                        a_rf, b_rf = tuple(rdcd_a.elements()), b_type(rdcd_b.elements())
                        a_perms = permutations(a_rf)

                except TypeError:
                    # TODO: We could probably get more coverage for this case
                    # by using `HashableForm`.
                    pass

                # If they're both ground and we're using basic unification,
                # then simply check that one is a permutation of the other and
                # be done.  No need to create and evaluate a bunch of goals in
                # order to do something that can be done right here.
                # Naturally, this assumes that the `isground` checks aren't
                # nearly as costly as all that other stuff.  If the gains
                # depend on the sizes of `a` and `b`, then we could do
                # `length_hint` checks first.
                if inner_eq == eq and isground(a_rf, S) and isground(b_rf, S):
                    if tuple(b_rf) in a_perms:
                        yield S
                        return
                    else:
                        # This has to be a definitive check, since we can only
                        # use the `a_perms` generator once; plus, we don't want
                        # to iterate over it more than once!
                        return

            yield from lany(inner_eq(b_rf, a_type(i)) for i in a_perms)(S)

        elif isinstance(b_rf, Sequence):

            b_type = type(b_rf)
            b_perms = permutations(b_rf)

            if no_ident:
                next(b_perms)

            yield from lany(inner_eq(a_rf, b_type(i)) for i in b_perms)(S)

        else:

            # None of the arguments are proper sequences, so state that one
            # should be and apply `permuteo` to that.

            a_itero_g = itero(
                a_rf, nullo_refs=(b_rf,), default_ConsNull=default_ConsNull
            )

            for S_new in a_itero_g(S):
                a_new = reify(a_rf, S_new)
                a_type = type(a_new)
                a_perms = permutations(a_new)

                if no_ident:
                    next(a_perms)

                yield from lany(inner_eq(b_rf, a_type(i)) for i in a_perms)(S_new)

    return permuteo_goal


# For backward compatibility
permuteq = permuteo
